#include <windows.h>
#include <stdio.h>
#include "memory.h"
#include "bcCommon.h"
#include "bcsvc.h"

static CRITICAL_SECTION InfoSynchro;
static struct InfoRec *Info=NULL;
static DWORD InfoCount=0;
static DWORD MaxError=20;
static DWORD ThreadCount=4;
static HANDLE StartEvent=NULL;
static BOOL AllowNetwork=0;

void InitInfo(void)
{
  HKEY hKey; DWORD Type,DataSize;
  if((RegOpenKeyEx(HKEY_LOCAL_MACHINE,"SYSTEM\\CurrentControlSet\\Services\\"SVC_NAME"\\Parameters",0,KEY_QUERY_VALUE,&hKey))==ERROR_SUCCESS)
  {
    DataSize=sizeof(MaxError);
    RegQueryValueEx(hKey,"MaxError",0,&Type,(LPBYTE)&MaxError,&DataSize);
    DataSize=sizeof(ThreadCount);
    RegQueryValueEx(hKey,"ThreadCount",0,&Type,(LPBYTE)&ThreadCount,&DataSize);
    if((ThreadCount>16)||(ThreadCount<1)) ThreadCount=4;
    DataSize=sizeof(AllowNetwork);
    RegQueryValueEx(hKey,"AllowNetwork",0,&Type,(LPBYTE)&AllowNetwork,&DataSize);
    RegCloseKey(hKey);
  }
  StartEvent=CreateEvent(NULL,TRUE,FALSE,NULL);
  InitializeCriticalSection(&InfoSynchro);
}

void FreeInfo(void)
{
  ResetStartEvent();
  SetAllStopInfo();
  WaitForInfo();
  DeleteCriticalSection(&InfoSynchro);
  if(Info) free(Info);
  Info=NULL; InfoCount=0;
}

void ResetStartEvent(void)
{
  if(StartEvent)
  {
    SetEvent(StartEvent);
    CloseHandle(StartEvent);
    StartEvent=NULL;
  }
}

void WaitStartEvent(void)
{
  if(StartEvent)
    while(WaitForSingleObject(StartEvent,3000)!=WAIT_OBJECT_0)
      ReportStatusToSCMgr(SERVICE_START_PENDING,NO_ERROR,3000);
}

void AddInfo(DWORD type,const wchar_t *Src,const wchar_t *SrcDir,const wchar_t *DestDir,BOOL InfoEx)
{
  DWORD ThreadId=GetCurrentThreadId();
  struct InfoRec *NewInfo;
  EnterCriticalSection(&InfoSynchro);
  NewInfo=(struct InfoRec *)realloc(Info,sizeof(struct InfoRec)*(InfoCount+1));
  if(NewInfo)
  {
    Info=NewInfo;
    Info[InfoCount].info.ThreadId=ThreadId;
    Info[InfoCount].info.type=type;
    wcscpy(Info[InfoCount].info.Src,Src);
    wcscpy(Info[InfoCount].info.SrcDir,SrcDir);
    wcscpy(Info[InfoCount].info.DestDir,DestDir);
    { //remove file name from DestDir
      wchar_t *fileptr=wcsrchr(Info[InfoCount].info.DestDir,L'\\');
      if(fileptr) fileptr[1]=0;
    }
    Info[InfoCount].info.percent=0;
    Info[InfoCount].info.InfoEx=InfoEx;
    wcscpy(Info[InfoCount].Src,L"");
    wcscpy(Info[InfoCount].Dest,L"");
    Info[InfoCount].stop=FALSE;
    Info[InfoCount].info.pause=FALSE;
    Info[InfoCount].TotalSize=0;
    Info[InfoCount].CurrentSize=0;
    Info[InfoCount].CurrentSizeAdd=0;
    Info[InfoCount].Errors=0;
    Info[InfoCount].PauseTime=0;
    Info[InfoCount].SizeType=0;
    InfoCount++;
  }
  else
  {
    free(Info);
    Info=NULL;
    InfoCount=0;
  }
  LeaveCriticalSection(&InfoSynchro);
}

void DelInfo(void)
{
  unsigned int i; DWORD ThreadId=GetCurrentThreadId();
  struct InfoRec *NewInfo;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      memmove(Info+i,Info+i+1,sizeof(struct InfoRec)*(InfoCount-i-1));
      NewInfo=(struct InfoRec *)realloc(Info,sizeof(struct InfoRec)*(InfoCount-1));
      if(NewInfo||(InfoCount==1))
      {
        Info=NewInfo;
        InfoCount--;
      }
      else
      {
        free(Info);
        Info=NULL;
        InfoCount=0;
      }
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
}

DWORD GetAllInfo(struct SmallInfoRec **GetInfo)
{
  unsigned int i; DWORD Result=0;
  EnterCriticalSection(&InfoSynchro);
  if(InfoCount)
  {
    *GetInfo=(struct SmallInfoRec *)malloc(sizeof(struct SmallInfoRec)*InfoCount);
    if(*GetInfo)
    {
      Result=InfoCount;
      for(i=0;i<InfoCount;i++)
        (*GetInfo)[i]=Info[i].info;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
  return Result;
}

void GetInfo(DWORD ThreadId,struct InfoRec *GetInfo)
{
  unsigned int i;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      *GetInfo=Info[i];
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
}

static void CheckPause(BOOL *Stop)
{
  if(!*Stop)
  {
    unsigned int i; DWORD ThreadId=GetCurrentThreadId(); BOOL wait=TRUE;
    while(wait)
    {
      wait=FALSE;
      EnterCriticalSection(&InfoSynchro);
      for(i=0;i<InfoCount;i++)
      {
        if(Info[i].info.ThreadId==ThreadId)
        {
          *Stop=Info[i].stop;
          wait=Info[i].info.pause&(!Info[i].stop);
          Info[i].PauseTime+=2000000ULL;
          break;
        }
      }
      LeaveCriticalSection(&InfoSynchro);
      Sleep(200);
    }
  }
}

BOOL UpdateInfo(wchar_t *newSrc,wchar_t *newDest)
{
  unsigned int i; DWORD ThreadId=GetCurrentThreadId(); BOOL Result=FALSE,Pause=FALSE;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      Result=Info[i].stop;
      Pause=Info[i].info.pause;
      if(!Result)
      {
        wcscpy(Info[i].Src,newSrc);
        wcscpy(Info[i].Dest,newDest);
      }
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
  if(Pause) CheckPause(&Result);
  return Result;
}

void SetStopInfo(DWORD ThreadId)
{
  unsigned int i;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      Info[i].stop=TRUE;
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
}

void SetPauseInfo(DWORD ThreadId)
{
  unsigned int i;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      Info[i].info.pause=!Info[i].info.pause;
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
}

BOOL SetStartInfo(unsigned long long size,unsigned int type)
{
  unsigned int i; DWORD ThreadId=GetCurrentThreadId(); BOOL Result=FALSE,Pause=FALSE;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      GetLocalTime(&Info[i].StartTime);
      Info[i].PauseTime=0;
      Info[i].TotalSize=size;
      Info[i].SizeType=type;
      Result=Info[i].stop;
      Pause=Info[i].info.pause;
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
  if(Pause) CheckPause(&Result);
  return Result;
}

BOOL UpdatePosInfo(unsigned long long add,unsigned long long subadd)
{
  unsigned int i; DWORD ThreadId=GetCurrentThreadId(); BOOL Result=FALSE,Pause=FALSE; double percent;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      if(Info[i].SizeType)
        Info[i].CurrentSize++;
      else
      {
        Info[i].CurrentSize+=add;
        Info[i].CurrentSizeAdd=subadd;
      }
      if(Info[i].TotalSize>0)
      {
        if(Info[i].TotalSize>=(Info[i].CurrentSize+Info[i].CurrentSizeAdd))
        {
          percent=(Info[i].CurrentSize+Info[i].CurrentSizeAdd);
          percent=percent/Info[i].TotalSize*100;
          Info[i].info.percent=percent;
        }
        else
          Info[i].info.percent=100;
      }
      Result=Info[i].stop;
      Pause=Info[i].info.pause;
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
  if(Pause) CheckPause(&Result);
  return Result;
}

DWORD CheckErrorInfo(void)
{
  unsigned int i; DWORD ThreadId=GetCurrentThreadId(); DWORD Result=0;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
  {
    if(Info[i].info.ThreadId==ThreadId)
    {
      Info[i].Errors++;
      if(Info[i].Errors==MaxError) Result=1;
      if(Info[i].Errors>MaxError) Result=2;
      break;
    }
  }
  LeaveCriticalSection(&InfoSynchro);
  return Result;
}

void SetAllStopInfo(void)
{
  unsigned int i;
  EnterCriticalSection(&InfoSynchro);
  for(i=0;i<InfoCount;i++)
    Info[i].stop=TRUE;
  LeaveCriticalSection(&InfoSynchro);
}

void WaitForInfo(void)
{
  BOOL wait=TRUE;
  while(wait)
  {
    EnterCriticalSection(&InfoSynchro);
    wait=InfoCount;
    LeaveCriticalSection(&InfoSynchro);
    Sleep(200);
  }
}

DWORD GetMaxError(void)
{
  return MaxError;
}

DWORD GetThreadCount(void)
{
  return ThreadCount;
}

BOOL GetAllowNetwork(void)
{
  return AllowNetwork;
}
